-- unitUtil.lua
-- by KirConjurer (angrykirc)

-- Some information
NOX_UNIT_FLAGS = {
    BELOW = 0x1, -- object does not collide with airborne units
    NO_UPDATE = 0x2, -- O is ignored while executing update queue
    ACTIVE = 0x4, -- UNKNOWN
    ALLOW_OVERLAP = 0x8, -- O will not be pushed upon collision with obstacles (effects, decorations)
    SHORT = 0x10, -- Jumping PC and airborne objects will not collide with O
    DESTROYED = 0x20, -- O is going to be removed next frame
    NO_COLLIDE = 0x40, -- O ignores collisions with anything (static decorations)
    MISSILE_HIT = 0x80, -- O triggers missile collisions
    EQUIPPED = 0x100, -- O is being worn by NPC or PC, any armor/weapons for example
    PARTITIONED = 0x200, -- UNKNOWN
    NO_COLLIDE_OWNER = 0x400, -- O does not collide with its parent
    OWNER_VISIBLE = 0x800, -- O is only visible to its parent (TeleportGlyphX)
    EDIT_VISIBLE = 0x1000, -- O is only visible in NoxEdit
    -- O with OWNER_VISIBLE / EDIT_VISIBLE will not render at all, even with show extents mode enabled
    NO_PUSH_CHARACTERS = 0x2000, -- O does not collide with PC
    AIRBORNE = 0x4000, -- O ignores collision with some objects
    DEAD = 0x8000, -- O is a monster/PC and is dead; used for various tests
    SHADOW = 0x10000, -- O blocks PC's line of sight; examples ColumnX, StoneBlock, Boulder
    FALLING = 0x20000, -- UNKNOWN, probably related to Pits
    IN_HOLE = 0x40000, -- UNKNOWN, probably related to Pits
    RESPAWN = 0x80000, -- O is added in respawn list upon creation (only in MP)
    ON_OBJECT = 0x100000, -- fall logic related - prevents smth
    SIGHT_DESTROY = 0x200000, -- UNKNOWN
    TRANSIENT = 0x400000, -- decaying object
    BOUNCY = 0x800000, -- obj has special fall logic
    ENABLED = 0x1000000, -- O is in "active" state (valid only for interactive objects - elevators, buttons...)
    PENDING = 0x2000000, -- set on recently created O's, probably forces MSG_SIMPLE_OBJECT
    TRANSLUCENT = 0x4000000, -- UNKNOWN, probably something related to lighting system
    STILL = 0x8000000, -- UNKNOWN
    NO_AUTO_DROP = 0x10000000, -- O does not drop upon death of its current holder; example OblivionOrb
    FLICKER = 0x20000000, -- UNKNOWN, something related to projectiles
    SELECTED = 0x40000000,
    MARKED = 0x80000000
}

NOX_CLASS_FLAGS = {
    MISSILE = 0x1, -- O is treated as projectile (some special update handling)
    MONSTER = 0x2, -- O is treated as monster, used mostly in validating
    PLAYER = 0x4, -- O represents player in game world. (NewPlayer, PhantomPlayer, Player, PlayerFemale)
    OBSTACLE = 0x8, -- O blocks movement, used mostly in monster AI code for static decorations
    FOOD = 0x10, -- O is a useful healing item, used in monster AI (ACTION_RETREAT)
    EXIT = 0x20, -- O is level finish (exit) point, examples: ExitX, TeleporterExit, PitExit
    KEY = 0x40, -- O is a key used for opening locked doors
    DOOR = 0x80, -- Object is a door
    INFO_BOOK = 0x100, -- O is a spell book or field guide
    TRIGGER = 0x200, -- O executes some event upon interaction (Buttons, levers, pressure plates, wells)
    TRANSPORTER = 0x400, -- O is used for instant travelling (TeleportPentagram, TeleportExitStoneX)
    HOLE = 0x800, -- PitCrumblingFloor / ForestPit
    WAND = 0x1000, -- O is a weapon for magic users (InfinitePainWand, LesserFireballWand, ForceWand...)
    FIRE = 0x2000, -- O is an instance of fire (Flame, FlameCleanse, BlueFlame...)
    ELEVATOR = 0x4000, -- XElevator, never (?) used in code
    ELEVATOR_SHAFT = 0x8000, -- XElevatorPit
    DANGEROUS = 0x10000, -- Marks O as dangerous for AI, used for pathfinding
    MONSTERGENERATOR = 0x20000, -- XGenerator
    READABLE = 0x40000, -- Various signs, never used in code
    LIGHT = 0x80000, -- Object is a light source
    SIMPLE = 0x100000, -- Most objects have this in declaration. MSG_SIMPLE_OBJ contains netID and obj position
    COMPLEX = 0x200000, -- Like SIMPLE, but used only (?) for monsters; MSG_COMPLEX_OBJ also contains animation frame
    IMMOBILE = 0x400000, -- Used for static decorations. Objects with this class are only synchronized with clients by special packets (MSG_DOOR_ANGLE)
    VISIBLE_ENABLE = 0x800000, -- Object is only rendered if "ENABLED" flag set
    WEAPON = 0x1000000, -- 
    ARMOR = 0x2000000, -- 
    NOT_STACKABLE = 0x4000000, -- Multiple objects do not stack in inventory, any WEAPON/ARMOR has this
    TREASURE = 0x8000000, -- Object is used in Scavenger Hunt game mode
    FLAG = 0x10000000, -- Object is a flag used in CTF game mode
    CLIENT_PERSIST = 0x20000000, -- 
    CLIENT_PREDICT = 0x40000000, --
    PICKUP = 0x80000000 --
}

NOX_WEAPON_FLAGS = {
    Flag = 0x1,
    Quiver = 0x2,
    Bow = 0x4,
    CrossBow = 0x8,
    RoundChakram = 0x40,
    FanChakram = 0x80,
    Sword = 0x100,
    Longsword = 0x200,
    GreatSword = 0x400,
    MorningStar = 0x800,
    BattleAxe = 0x1000,
    OgreAxe = 0x2000,
    WarHammer = 0x4000,
    StaffWooden = 0x8000,
    SulphorousFlareWand = 0x10000,
    SulphorousShowerWand = 0x20000,
    ForceWand = 0x40000,
    LesserFireballWand = 0x80000,
    DemonsBreathWand = 0x80000,
    -- Those two have equal masks...
    InfinitePainWand = 0x200000,
    DeathRayWand = 0x400000,
    OblivionHalberd = 0x800000,
    OblivionHeart = 0x1000000,
    OblivionWierdling = 0x2000000,
    OblivionOrb = 0x4000000
}

NOX_DAMAGE_TYPE = {
    BLADE = 0, -- weapons
    FLAME = 1, -- FlameX, Fireball
    CRUSH = 2, -- Bear, Golem
    IMPALE = 3, -- Halberd, Wasp
    DRAIN = 4, -- Ghost, Shade
    POISON = 5, -- Poison, Toxic cloud
    DISPEL_UNDEAD = 6, -- Dispel undead
    EXPLOSION = 7, -- BlackPowderBarrel explosions
    BITE = 8, -- Spiders, Wolves
    ELECTRIC = 9, -- Lightning?
    CLAW = 10, -- EmberDemon
    IMPACT = 11, -- Ogres, Troll, Hammer?
    LAVA = 12, -- Lava tiles
    DEATH_MAGIC = 13, -- Death spell
    PLASMA = 14, -- Plasma spell (Staff of Oblivion)
    MANA_BOMB = 15, -- Mana Bomb spell
    ZAP_RAY = 16, -- Death Ray spell
    AIRBORNE_ELECTRIC = 17 -- unknown
}

NOX_MONSTER_STATUS = {
    DESTROY_WHEN_DEAD = 1,
    CHECK = 2,
    CAN_BLOCK = 4,
    CAN_DODGE = 8,
    unused = 0x10,
    CAN_CAST_SPELLS = 0x20,
    HOLD_YOUR_GROUND = 0x40,
    SUMMONED = 0x80,
    ALERT = 0x100,
    INJURED = 0x200,
    CAN_SEE_FRIENDS = 0x400,
    CAN_HEAL_SELF = 0x800,
    CAN_HEAL_OTHERS = 0x1000,
    CAN_RUN = 0x2000,
    RUNNING = 0x4000,
    ALWAYS_RUN = 0x8000,
    NEVER_RUN = 0x10000,
    BOT = 0x20000,
    MORPHED = 0x40000,
    STAY_DEAD = 0x80000,
    ON_FIRE = 0x100000,
    FRUSTRATED = 0x200000
}

local intToPtr;
do
	local t = memAlloc(4)
	function intToPtr(num)
        	setPtrInt(t, 0, num)
        	return getPtrPtr(t, 0)
    	end
end -- thanks to Dio

local fltToPtr;
do
	local t = memAlloc(4)
	function fltToPtr(num)
        	setPtrFloat(t, 0, num)
        	return getPtrPtr(t, 0)
    	end
end

local oneUserdata = intToPtr(1)
-- SUPER HACK BEGIN
setPtrInt(intToPtr(0x6F7B20), 0, 0)
local zeroUserdata = ptrCall2(0x4896C0, oneUserdata, oneUserdata)
-- SUPER HACK END
local noxCosTable = intToPtr(0x5B6658)
local noxSinTable = intToPtr(0x5B665C)
local float255 = intToPtr(0x437F0000)
local gameFPSAddr = intToPtr(0x85B3FC)

hostPlayer = getPtrPtr(ptrCall2(0x417090, intToPtr(31), oneUserdata), 0x808)
    
function unitOnUpdateHack(unit, fn)
    local oldHandler = getPtrPtr(unit, 0x2D4)
    unitOnDie(unit, fn)
    local newHandler = getPtrPtr(unit, 0x2D4)
    if (oldHandler) then
        setPtrPtr(unit, 0x2D4, oldHandler)
    else
        setPtrInt(unit, 0x2D4, 0)
    end -- is this required? after 1 frame it will be reset in original handler...
    setPtrPtr(unit, 0x2E8, newHandler)
end

function unitAnimInit(unit, smth)
    ptrCall2(0x4E4800, unit, intToPtr(smth))
end

function playerSetAnim(unit, anim)
    ptrCall2(0x4FA020, unit, intToPtr(anim))
    --[[local uc = getPtrPtr(unit, 0x2EC)
    setPtrByte(uc, 0x59, getPtrByte(uc, 0x58, anim))
    setPtrByte(uc, 0x58, anim)
    setPtrByte(uc, 0xEC, 0)]]--
end

function netSendMessage(unit, text)
    return ptrCall2(0x4DA2C0, unit, text)
    --return ptrCall2(0x4D9FD0, intToPtr(plrid), text) -- 3 arg bro...
end

function unitIsMonster(unit)
    if (bitAnd(getUnitClass(unit), 2) == 2 and bitAnd(unitFlags(unit), 0x8000) ~= 0x8000) then return true end
    return false -- true only if unit class is MONSTER and flag DEAD is not set
end

function unitCallUseBy(user, unit)
    return ptrCall2(0x53F8E0, user, unit)
end

function monsterClearActionStack(unit)
    ptrCall2(0x50A3A0, unit, unit) -- clear action stack
    return true
end

function monsterGetCurrentAction(unit)
    local uc = getPtrPtr(unit, 0x2EC)
    local topAction = getPtrInt(uc, 0x228)
    return topAction
end

function monsterActionIsMoveTo(unit)
    if monsterGetCurrentAction(unit) == 7 then return true end
    return false
end

function monsterActionIsDodge(unit)
    if monsterGetCurrentAction(unit) == 9 then return true end
    return false
end

function unitIsPlayer(unit)
    return bitAnd(getUnitClass(unit), 4) == 4
end

function unitIsDead(unit)
    if not objectExists(unit) then return true end
    return bitAnd(unitFlags(unit), 0x8000) == 0x8000
end

function unitCountCtrlMonsters(unit)
    local nextSlave = getPtrPtr(unit, 0x204)
    local slave = 0
    while objectExists(nextSlave) do
        if unitIsMonster(nextSlave) then slave = slave + 1 end
        nextSlave = getPtrPtr(nextSlave, 0x200)
    end
    return slave
end

function unitNoSpellFistActive(unit)
    local nextSlave = getPtrPtr(unit, 0x204)
    local fovActive = false
    while objectExists(nextSlave) do
        if getThingName(nextSlave) == "LargeFist" then fovActive = true end
        nextSlave = getPtrPtr(nextSlave, 0x200)
    end
    return not fovActive
end

function unitFindParentPlayer(unit)
    return ptrCall2(0x4EC580, unit, unit)
end

function utilGetGameFPS()
    return getPtrInt(gameFPSAddr, 0)
end

function utilSecToFrames(seconds, multiplier)
    local val = utilGetGameFPS()
    val = val * seconds * multiplier
    return math.ceil(val)
end

local function memFree2(mem)
    ptrCall2(0x40425D, mem, mem)
end

function playerAddGold(plr, gold)
    return ptrCall2(0x4FA590, plr, intToPtr(gold))
end

function unitGetParent(unit)
    return getPtrPtr(unit, 0x1FC)
end

-- including PlayerStart
function mapFindWaypointForBlink(unit)
    local point = memAlloc(8)
    setPtrFloat(point, 0, 0) setPtrFloat(point, 4, 0)
    ptrCall2(0x579F00, point, unit)
    local x = getPtrFloat(point, 0)
    local y = getPtrFloat(point, 4)
    memFree2(point)
    return x, y
end

function netGreenLightning(x1, y1, x2, y2, etime)
    local rectStruct = memAlloc(16)
    setPtrInt(rectStruct, 0, x1)
    setPtrInt(rectStruct, 4, y1)
    setPtrInt(rectStruct, 8, x2)
    setPtrInt(rectStruct, 12, y2)
    ptrCall2(0x523790, rectStruct, intToPtr(etime))
    memFree2(rectStruct)
end

function creatureActionRoam(unit)
    local h = unitSetAction(unit, 10)
    setPtrInt(h, 4, 0)
    setPtrInt(h, 12, 0x80)
    return h
end

function netAllPlayersMonitorCreature(unit)
    local pid = 255--playerInfo(plr).idx
    if (pid) then
        ptrCall2(0x4D9250, intToPtr(pid), unit)
    else
        print("Omg this player is #0!..")
    end
end 

function netReportUnitHPAllPlayers(unit)
    local pid = 255
    ptrCall2(0x4D8620, intToPtr(pid), unit)
end

function unitLastDamageDiff(unit) -- frames
    local ldt = getPtrInt(unit, 0x218)
    return getFrameCounter() - ldt
end

function unitAddHP(unit, hp)
    return ptrCall2(0x4EE460, unit, intToPtr(hp))
end

function buffSetPower(unit, buff, pwr)
    setPtrByte(unit, 0x198 + buff, pwr)
    return unit
end

-- unitHP returns dword...
function unitGetHPSafe(unit)
    local hd = getPtrPtr(unit, 0x22C)
    return getPtrShort(hd, 0)
end

function unitTeleportWake(x, y, tx, ty, ftime)
    local t = createObject("TeleportWake", x, y)
    setPtrInt(t, 0x88, getFrameCounter() + ftime) 
    local c = memAlloc(8)
    setPtrFloat(c, 0, tx)
    setPtrFloat(c, 4, ty)
    setPtrPtr(t, 0x2BC, c)
    return t
end

function unitAllocNewHP(unit, hp)
    local hd = memAlloc(0x14)
    setPtrShort(hd, 0, hp)
    setPtrShort(hd, 2, hp)
    setPtrShort(hd, 4, hp)
    setPtrPtr(unit, 0x22C, hd)
    return hd
end

function unitGetMaxHP(unit)
    local hd = getPtrPtr(unit, 0x22C)
    return getPtrShort(hd, 4)
end

function unitSetCurrentHP(unit, val)
    if val < 1 then val = 1 end
    ptrCall2(0x4E4560, unit, intToPtr(val))
end

function unitSetMaxHealth(npc, val)
    local hd = getPtrPtr(npc, 0x22C)
    setPtrShort(hd, 0, val)
    setPtrShort(hd, 2, val)
    setPtrShort(hd, 4, val)
    return
end

function unitAddTemporaryHP(unit, hp, leng)
    local health1 = unitGetHPSafe(unit)
    local health2 = unitGetMaxHP(unit)
    local restoreRealHP = function()
        if not unitIsDead(unit) then
            health1 = health2 - unitGetHPSafe(unit)
            if health1 > 0 then health1 = 0 end
            health1 = unitGetHPSafe(unit) + health1
            unitSetMaxHealth(unit, health2)
            unitSetCurrentHP(unit, health1)
        end
    end
    unitSetMaxHealth(unit, health2 + hp)
    unitSetCurrentHP(unit, health1 + hp)
    ptrCall2(0x4EE4C0, unit, unit) -- inform owner
    setTimeout(restoreRealHP, leng)
end

-- sets/unsets NO_COLLIDE flag
function unitNoCollide(unit, collide)
    if (collide) then setPtrInt(unit, 0x10, bitXor(unitFlags(unit), 0x40))
    else setPtrInt(unit, 0x10, bitOr(unitFlags(unit), 0x40)) end
    return collide
end

function objectExists(unit)
    --if unit == zeroUserdata then return false end
    if not unit then return false end
    if getThingType(unit) == 0xACAC then return false end
    return true
end

-- sets/unsets AIRBORNE flag
function unitFlAirborne(unit, flag)
    if (flag) then setPtrInt(unit, 0x10, bitXor(unitFlags(unit), 0x4000))
    else setPtrInt(unit, 0x10, bitOr(unitFlags(unit), 0x4000)) end
    return flag
end

function createHarpoon(unit)
    -- works only for players
    return ptrCall2(0x538890, unit, oneUserdata)
end

function memLoadString8B( str ) -- by Dio
 local size = #str
 local mem = memAlloc( size )
 setPtrByte( mem, size, 0 )
 for i = 1, size do
  setPtrByte( mem, i-1, string.byte( str, i ) )
 end
 return mem
end

function memLoadString16B( str )
    local size = #str
    size = size * 2
    local mem = memAlloc( size )
    local strNdx = 1
    setPtrShort( mem, size, 0 )
    for i = 0, size - 2, 2 do
        setPtrShort( mem, i, string.byte( str, strNdx ) )
        strNdx = strNdx + 1
    end
    return mem
end

-- temp fix (server only) for netClientPrint.
function conColorPrint(text, color)
    if (color == nil) or (color == 0) then
        color = 4
    end
    local str = memLoadString16B(text) -- *wchar_t
    ptrCall2(0x450B90, intToPtr(color), str)
    memFree2(str) -- freeing temporary memory
    -- string gets copied to another buffer
end

function findAudioID(audname)
    local str = memLoadString8B(audname)
    return ptrCall2(0x40AF50, str, oneUserdata)
end

function monsterGetSoundset(unit)
    return ptrCall2(0x424300, unit, unit)
end

function unitCanSeeUnit(unitA, unitB)
    local ax, ay = unitPos(unitA)
    local bx, by = unitPos(unitB)
    if mapTraceRay(ax, ay, bx, by, 1) then
        if not (unitTestBuff(unitB, 0) and not unitTestBuff(unitA, 21)) then return true end
         -- also check for invisibility
    end
    return false
end

function unitGetDirection(unit)
    return getPtrShort(unit, 0x7C)
end

function unitGetLookAngle(unit)
    return getPtrShort(unit, 0x7E)
end

function arrowSetDamage(unit, dmg) -- works only for MonsterArrowCollide
    local collData = getPtrPtr(unit, 0x2BC)
    setPtrInt(collData, 0, dmg) -- SOLO
    setPtrInt(collData, 4, dmg) -- ARENA
    return unit
end

-- unitBecomePet for non-monsters
function unitSetParent(unitA, unitB)
    return ptrCall2(0x4EC290, unitB, unitA)
end

function unitClearParent(unit)
    return ptrCall2(0x4EC300, unit, unit)
end

function unitNeedSync(unit)
    return ptrCall2(0x4E44F0, unit, unit)
end

function unitIsPoisonImmune(unit)
    local subclass = getPtrInt(unit, 0xC)
    return bitAnd(subclass, 0x200) == 0x200
end

function unitNoPoisonResist(unit)
    local subclass = getPtrInt(unit, 0xC)
    if bitAnd(subclass, 0x200) == 0x200 then
        -- IMMUNE_POISON
        setPtrInt(unit, 0xC, bitAnd(subclass, 0x200))
    end
    return unit
end

function unitSetOwnerVisible(unit, only) -- hides unit from sight
    local flags = unitFlags(unit)
    if (only) then
        setPtrInt(unit, 0x10, bitOr(flags, 0x800)) -- OWNER_VISIBLE
    elseif bitAnd(flags, 0x800) == 0x800 then
        setPtrInt(unit, 0x10, bitXor(flags, 0x800))
    end
end

function unitSetLookAngle(unit, val)
    setPtrShort(unit, 0x7C, val)
    setPtrShort(unit, 0x7E, val)
    return unit
end

function utilCosFromDirect(dir)
    if (dir < 0) then dir = 255 + dir end
    if (dir > 255) then dir = dir - 255 end
    return getPtrFloat(noxCosTable, 2 * dir * 4)
end

function utilSinFromDirect(dir)
    if (dir < 0) then dir = 255 + dir end
    if (dir > 255) then dir = dir - 255 end
    return getPtrFloat(noxSinTable, 2 * dir * 4)
end

function playerGetName(plr)
    if (not plr) then return "<404>" end
    return playerInfo(plr).name
end
    
function netReportDequip(item)
    return ptrCall2(0x4D84C0, intToPtr(255), item)
end

function monsterNoSpells(monster)
    for i = 0, 136 do
        monsterSetSpell(monster, i, 0)
    end
end

function monsterSetSpell(npc, spell, useflag)
    local uc = getPtrPtr(npc, 0x2EC)
    local offset = 4 * spell + 0x5D0
    setPtrInt(uc, offset, useflag)
    return
end

function unitIsCharmable(unit)
    local tt = intToPtr(getThingType(unit))
    local gd = ptrCall2(0x4272B0, tt, tt)
    local tmp = ptrCall2(0x427380, gd, gd)
    if tmp == oneUserdata then return true end
    return false
end

-- returns true if unit is not null and active (enabled)
function unitIsOn(unit)
    if (not unit) then return false end
    if (bitAnd(unitFlags(unit), 0x1000000) ~= 0x1000000) then return false end
    return true
end

function playerSubGold(plr, gold)
    return ptrCall2(0x4FA5D0, plr, intToPtr(gold))
end

function buffCancel(unit, buff) -- HACK
    local buffUD = zeroUserdata
    if (buff > 0) then buffUD = intToPtr(buff) end
    return ptrCall2(0x4FF5B0, unit, buffUD)
end

function playerGetGold(plr)
    local uc = getPtrPtr(plr, 0x2EC)
    local pi = getPtrPtr(uc, 0x114)
    return getPtrInt(pi, 0x874)
end

function netArrowTrapFX(x, y, n)
    local p = memAlloc(8) -- point struct
    setPtrFloat(p, 0, x)
    setPtrFloat(p, 4, y)
    ptrCall2(0x5238A0, p, intToPtr(n))
    memFree2(p)
    return true
end

function netGeneratorBreakFX(x, y)
    local p = memAlloc(8) -- point struct
    setPtrFloat(p, 0, x)
    setPtrFloat(p, 4, y)
    ptrCall2(0x523200, p, intToPtr(0xC8))
    memFree2(p)
    return true
end

function netSendEarthquake(x, y, power)
    local p = memAlloc(8) -- point struct
    setPtrFloat(p, 0, x)
    setPtrFloat(p, 4, y)
    ptrCall2(0x4D9110, p, intToPtr(power))
    memFree2(p)
    return true
end

function mobCanMove(unit)
    return not (unitTestBuff(unit, 5) or unitTestBuff(unit, 25) or unitTestBuff(unit, 3) or unitTestBuff(unit, 11))
end

function createMagicMissile(x, y, parent)
    if (not parent) then return nil end
    local m = createObject("MagicMissile", x, y)
    local uc = getPtrPtr(m, 0x2EC)
    setPtrPtr(uc, 0, parent)
    unitSetParent(m, parent)
    return m
end

function createFallingMeteor(x, y, fallslow, damage)
    local m = createObject("Meteor", x, y)
    local uc = getPtrPtr(m, 0x2EC)
    setPtrInt(uc, 0, damage) -- meteor max damage
    setPtrInt(m, 0x14, bitOr(getPtrInt(m, 0x14), 0x20)) -- dunno
    ptrCall2(0x4E46F0, m, float255) -- unitRaise
    setPtrFloat(m, 0x6C, fallslow) -- negative values make falling faster, positive - slower
    return m
end

function createToxicCloud(x, y, lifetime)
    local tc = createObject("ToxicCloud", x, y)
    local uc = getPtrPtr(tc, 0x2EC)
    setPtrInt(uc, 0, lifetime)
    return tc
end

local function getSomeRewardData()
    return ptrCall2(0x4E3CC0, oneUserdata, oneUserdata)
end

function activateRewardMarker(marker)
    return ptrCall2(0x4F0720, marker, getSomeRewardData())
end

function unitSetOn(unit)
	return ptrCall2(0x4E75B0, unit, oneUserdata)
end

function unitSetOff(unit)
	return ptrCall2(0x4E7600, unit, oneUserdata)
end

function unitCanAttack(unitA, unitB)
    local p = ptrCall2(0x5330C0, unitA, unitB)
    if (p == oneUserdata) then return true end
    return false
end

function unitSetTeam(unit, team)
    if (unit) then
        setPtrInt(unit, 0x34, team)
        --setTimeout(ptrCall2(0x004E44F0, unit, unit)
    end
end

function createTreasure(x, y)
    local tre = createObject("TreasurePile", x, y)
    setPtrInt(tre, 0x2C4, 0x4F3580) -- TreasurePickup
    setPtrInt(tre, 0x2C8, 0x4ED710) -- TreasureDrop
    return tre
end -- use with gameFlags(bitOr(gameFlags(), 64))

function unitGetTeam(unit)
    if (unit) then
        return getPtrInt(unit, 0x34)
    end
end

function servPatchDodgeAnimation()
    local addr = intToPtr(0x5338B0)
    if getPtrInt(addr, 0) == 0x5337E5 then
        setPtrInt(addr, 0, 0x5337CE)
    end
end

function unitDamagePure(unit, val)
    setPtrInt(unit, 0x218, getFrameCounter()) -- damage taken
    return ptrCall2(0x4EE5E0, unit, intToPtr(val))
end

function unitReanimate(unit)
    if unitIsDead(unit) then
        unitActionPop(unit)
        unitActionPop(unit)
        unitSetAction(unit, 0) -- ACTION_IDLE
        local f = getPtrInt(unit, 0x10)
        f = bitXor(f, 0x8000) -- DEAD
        if (bitAnd(f, 0x18) == 0x18) then
            f = bitXor(f, 0x18) -- NO_COLLIDE|SHORT
        end
        setPtrInt(unit, 0x10, f)
        ptrCall2(0x4EE6F0, unit, unit) -- restore health
        ptrCall2(0x4DA8D0, unit, unit) -- re-add to updatable
        setPtrInt(unit, 0x2E8, 0x50A5C0) -- restore update fn
    end
end

function unitIsPoisoned(unit)
    local result = getPtrByte(unit, 0x21C)
    return result ~= 0
end

function unitPoison(unit, lvl)
    if (unit) then
        if (lvl > 0) then
            soundMake(100, unitPos(unit))
            setPtrByte(unit, 0x21C, lvl)
            --[[setPtrShort(unit, 0x21E, 1000)
            local hpData = getPtrPtr(unit, 0x22C)
            setPtrInt(hpData, 0x10, getFrameCounter())
            if (bitAnd(getUnitClass(unit), 4)) then
                local uc = getPtrPtr(unit, 0x2EC)
                local pd = getPtrPtr(uc, 0x114)
                local fl = getPtrInt(pd, 0xE60)
                setPtrInt(pd, 0xE60, bitOr(fl, 0x400))
            end]]--
        else
            ptrCall2(0x4EE9D0, unit, unit)
            --setPtrByte(unit, 0x21C, 0)
            --setPtrShort(unit, 0x21E, 0)
        end
        ptrCall2(0x4EEA90, unit, intToPtr(lvl)) 
        -- temporary hack: call original function which sends update packet
    end
end

-- Example - translated from Nox asm code
-- local X, Y = unitPos(barrel)
-- unitSplashDamage(unit, X, Y, 100.0, 30.0, 30, 7, barrel, 0)
function unitSplashDamage(unit, x, y, maxDistance, fadeDistance, damage, dmgType, parent1)
    local pointPtr = memAlloc(8)
    setPtrFloat(pointPtr, 0, x) 
    setPtrFloat(pointPtr, 4, y)
    local structPtr = memAlloc(0x1C)
    setPtrPtr(structPtr, 0, pointPtr)
    setPtrFloat(structPtr, 4, maxDistance)
    setPtrFloat(structPtr, 8, fadeDistance)
    setPtrInt(structPtr, 0xC, damage)
    setPtrInt(structPtr, 0x10, dmgType)
    if (parent1) then setPtrPtr(structPtr, 0x14, parent1)
    else setPtrInt(structPtr, 0x14, 0) end
    setPtrInt(structPtr, 0x18, 0) -- ignore parent2
    ptrCall2(0x4E2700, unit, structPtr)
    memFree2(pointPtr)
    memFree2(structPtr)
end

function unitDamageHack(unit, dmgType, damage, parent)
    local x, y = unitPos(unit)
    unitSplashDamage(unit, x, y, 4, 4, damage, dmgType, parent)
end

function playMusicHack(musNum, volume)
    if musNum > 29 then return false end
    if musNum < 1 then 
        ptrCall2(0x516520, oneUserdata, oneUserdata)
    else
        ptrCall2(0x507230, intToPtr(musNum), oneUserdata)
        ptrCall2(0x507230, intToPtr(volume), oneUserdata)
        ptrCall2(0x516430, oneUserdata, oneUserdata)
    end
    return true
end

function unitAddRecentlyCreated()
    ptrCall2(0x4DAC00, oneUserdata, oneUserdata)
    return true
end

function monsterPickMeleeStrike(monster)
    local targ = ptrCall2(0x549440, monster, zeroUserdata)
    if (targ == zeroUserdata) then return nil end
    return targ
end

function monsterGetAnimFrame(monster)
    local uc = getPtrPtr(monster, 0x2EC)
    return getPtrByte(uc, 0x1E1)
end

function monsterSetAnimFrame(monster, frame)
    local uc = getPtrPtr(monster, 0x2EC)
    setPtrByte(uc, 0x1E1, frame)
end

function monsterIsAttacking(monster)
    --[[local uc = getPtrPtr(monster, 0x2EC)
    local fr = getPtrByte(uc, 0x1E2)
    return fr == 0]]--
    if monsterGetCurrentAction(unit) == 16 then return true end
    return false
end

local function bshl(a, b)
    return a * 2 ^ b
end

function buffApplySilent(unit, buffN, duration, power)
    if bitAnd(unitFlags(unit), 0x8022) > 0 then return false end
    setPtrShort(unit, 0x158 + 2 * buffN, duration)
    if (not power) then power = 3 end
    setPtrByte(unit, 0x198 + buffN, power)
    local flags = bitOr(bshl(1, buffN), getPtrInt(unit, 0x154))
    if (flags == 0) then ptrCall2(0x4E48F0, unit, zeroUserdata)
    else ptrCall2(0x4E48F0, unit, intToPtr(flags)) end
    return true
end

function cursorSetIcon(num)
    local ud = zeroUserdata
    if (num > 0) then ud = intToPtr(num) end
    ptrCall2(0x477610, ud, ud)
end

function playerGetStrength(plr)
    local plri = getPtrPtr(getPtrPtr(plr, 0x2EC), 0x114)
    return getPtrInt(plri, 0x8BF)
end

function playerSetStrength(plr, val)
    local plri = getPtrPtr(getPtrPtr(plr, 0x2EC), 0x114)
    setPtrInt(plri, 0x8BF, val)
end

function unitNoDeathDrop(unit)
    setPtrInt(unit, 0x10, bitOr(getPtrInt(unit, 0x10), 0x10000000))
end

function monsterRegisterFireSprite()
    local ptrToFirstDef = intToPtr(0x81B180)
    local firstDef = getPtrPtr(ptrToFirstDef, 0)
    local thingid = getThingType("FireSprite")
    if (getPtrInt(firstDef, 0xF0) ~= thingid) then
        local ptr = memAlloc(0xF8)
        setPtrInt(ptr, 0x44, 50) -- HEALTH
        setPtrInt(ptr, 0x4C, 70) -- SPEED
        setPtrFloat(ptr, 0x60, 1) -- RUN_MULTIPLIER
        setPtrInt(ptr, 0x74, 15) -- MELEE_DAMAGE
        setPtrInt(ptr, 0x7C, 10) -- MELEE_ATTACK_DAMAGE_TYPE
        setPtrFloat(ptr, 0x78, 0) -- MELEE_ATTACK_IMPACT
        setPtrInt(ptr, 0x6C, 4) -- MELEE_ATTACK_FRAME
        setPtrInt(ptr, 0x80, 8) -- MELEE_ATTACK_MIN_DELAY
        setPtrInt(ptr, 0x84, 12) -- MELEE_ATTACK_MAX_DELAY
        setPtrFloat(ptr, 0x70, 25) -- MELEE_ATTACK_RANGE
        setPtrInt(ptr, 0xEC, 0x549380) -- MELEE_STRIKE_FUNCTION
        setPtrInt(ptr, 0xE4, 0) -- DIE_FUNCTION
        setPtrInt(ptr, 0xE8, 0) -- DEAD_FUNCTION
        setPtrInt(ptr, 0x5C, 0x10000) -- STATUS
        setPtrInt(ptr, 0xF0, thingid)
        setPtrPtr(ptr, 0xF4, firstDef)
        setPtrPtr(ptrToFirstDef, 0, ptr)
    end
end

--[[
useful for custom effects
example: netLinearOrbMove(unit, 5, 5, -0.01)
List of all "LinearOrbs":
DrainManaOrb
HealOrb
CharmOrb
WhiteOrb
WhiteVortexOrb
ManaBombOrb
WhiteMoveOrb
BlueMoveOrb
]]--
function netLinearOrbMove(unit, vecX, vecY, speed, lifetime)
    if (unit) then
        setPtrFloat(unit, 0x50, vecX)
        setPtrFloat(unit, 0x54, vecY)
        setPtrFloat(unit, 0x70, speed)
        ptrCall2(0x0523530, unit, unit)
        if (lifetime) then unitDecay(unit, lifetime) end
    end
end

function unitRaiseHack(unit, height)
    ptrCall2(0x4E46F0, unit, fltToPtr(height))
end

function unitToggleFlag(unit, fl)
    setPtrInt(unit, 0x10, bitXor(getPtrInt(unit, 0x10), fl))
end

-- used for lightning/manadrain/greater heal
function netStartDurationSpell(parent, spellN, target, power)
    local spellD = memAlloc(0x70)
    setPtrInt(spellD, 4, spellN)
    setPtrPtr(spellD, 0x30, target)
    setPtrInt(spellD, 8, power)
    if target then setPtrPtr(spellD, 0x10, parent)
    else setPtrInt(spellD, 0x10, 0) end
    --ptrCall2(0x4FEF90, spellD, target)
    ptrCall2(0x4FF130, spellD, spellD)
end

function netStopDurationSpell(parent, spell)
    ptrCall2(0x4FEB10, intToPtr(spell), parent)
end

-- modify object sell price
function unitSetWorth(obj, worth)
    setPtrInt(obj, 0x20, worth)
end
--[[function projectileReflectFrom(proj, unit)
    ptrCall2(0x4E0A70, 
    ]]--